# DEPRECATED
# This module is an alias to the network module and should not be used.
# It is temporarily left in place so that existing references do not break.
from .network import *

print("""
The 'networking_api_examples' module is deprecated! Use the 'network' module
directly going forward.
""")
